/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
 * 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010, 2011  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-lists.c,v 2.13 2011/10/03 14:14:30 acau Exp $
 */


#include "system.h"




/* FIXLIST e n */
/* VARLIST e n */
void reduce_list(e)
NODEPTR e;
{
    NODEPTR loc;
    int     cell_type,
            size_type;
    NODEPTR n = reduce_val (node_lnk_2(e));
    if (Ready) {
        type_check (n, INT);
        loc = reduce_ref (node_lnk_1(e));
        if (Ready) {
            cell_type = is_static_mem (loc) ? STATIC : STATE;
            size_type = (node_type(e) == VARLIST) ? STATE : STATIC;
            assign_struct (loc, new_list1 (cell_type, size_type, n, node_mem_history(loc)));
            result (True);
        }
    }
}


/******************\
*                  *
*  list operators  *
*                  *
\******************/

void reduce_STRUCT_val(e)
NODEPTR e;
{
    NODEPTR e1;
    if (node_type(e1 = node_lnk_1(e)) == LIST)
        Ready = T;
    else {
        e1 = reduce_ref_and_fetch (e1);
        if (Ready)
            result (node1 (STRUCT, e1));
    }
}

void reduce_STRUCT_ref(e)
NODEPTR e;
{
    ref_result (reduce_ref (node_lnk_1(e)));
}

/* produce a static list */
/* MAKELIST elist */
void reduce_MAKELIST(e)
NODEPTR e;
{
    NODEPTR elist = node_lnk_1(e);
    NODEPTR item;
    unsigned int     size = elist_size (elist);
    NODEPTR blk = mem_block1 (STATIC, size + 1, 1);
    NODEPTR size_loc = node_blk_base(blk),
        base_loc = size_loc + offset2(1,1),
            loc = curr_loc (base_loc);

    Ready = T;
        /*printf("reduce_MAKELIST: base_loc=%p\n",base_loc);*/
    put_mem1 (size_loc, int_node ((int)size));

    for (elist = node_lnk_1(e); elist != NULL; elist = node_lnk_2(elist)) {
        item = reduce_val (node_lnk_1(elist));
        if (Ready) {
            put_mem1 (loc, copy_val (item));
            loc += offset2 (1,1);
        }
        else
            return;
    }
    if (Ready)
        result (list_node (blk, size_loc, base_loc));
}

int reduce_index(e)
NODEPTR e;
{
    e = reduce_val (e);
    int n;
    
    if (Ready) {
        type_check (e, INT);
        n = node_ival(e);
        assert0 (n>=0 , "index variable should be greater than zero");
        return (node_ival(e));
    }
    return (0);
}

/* SUB e1 e2 */
void reduce_SUB(e)
NODEPTR e;
{
    NODEPTR list;
    int     n = reduce_index (node_lnk_2(e));
    if (Ready) {
        list = reduce_ref_and_fetch (node_lnk_1(e));
        if (Ready) {
            switch (node_type(list)) {
                case LIST:
                    assert2d ((0 <= n) && (n < max_size (list)),
                             "subscript should be less than %d: %d",
			     max_size (list), n);
                        /*printf("reduce_SUB: base_loc=%p\n",list_base (list));*/
                    ref_result (list_base (list) + offset2 (n,node_mem_history(list_base (list))));
                    break;
                case STRING:
                    ref_result (make_ref (substring (list, n, n)));
                    break;
                default:
                    error_break1 ("cannot subscript expression of type %s",
                            print_name (list));
	    }
	}
    }
}

/* SIZE e, |e| */
void reduce_SIZE(e)
NODEPTR e;
{
    NODEPTR list = reduce_ref_and_fetch (node_lnk_1(e));
    if (Ready) {
        switch (node_type(list)) {
            case LIST:
                ref_result (list_size (list));
                break;
            case STRING:
                ref_result (make_ref (int_node (string_size (list))));
                break;
            default:
                error_break1 ("cannot get size of %s", print_name (list));
	}
    }
}


NODEPTR sublist(list,m,n)
NODEPTR list;
int m,n;
{
    int     sub_len, current_size;
    NODEPTR blk;
    NODEPTR size_loc, sub_base_loc;

    size_loc = list_size (list);
    current_size = get_int (size_loc);
    assert2d ((m >= 0) && (n <= current_size),
            "sublist [%d..%d] not within list", m, n);
    sub_len = (n > m) ? n - m : 0;
        /*printf("sublist: base_loc=%p\n",list_base (list));*/
    sub_base_loc = list_base (list) + offset2 (m,node_mem_history(size_loc));
        /*printf("sublist: sub_base_loc=%p\n",sub_base_loc);*/
    blk = local_mem1 (STATIC,node_mem_history(size_loc));
    size_loc = is_next (size_loc) ? next_loc (node_blk_base(blk)) :
        curr_loc (node_blk_base(blk));
    put_mem1 (size_loc, int_node (sub_len));
    return (list_node (blk, size_loc, sub_base_loc));
}

/* sublist of e from e1 to e2-1 */
/* SUBLIST e e1 e2 */
void reduce_SUBLIST(e)
NODEPTR e;
{
    NODEPTR list;
    int     m,
            n;
    m = reduce_index (node_lnk_2(e));
    if (Ready) {
        n = reduce_index (node_lnk_3(e));
        if (Ready) {
            list = reduce_ref_and_fetch (node_lnk_1(e));
            if (Ready) {
                switch (node_type(list)) {
                    case LIST:
                        if ((Ready = list_size_defined (list)))
                            ref_result (make_ref (sublist (list, m, n)));
                        break;
                    case STRING:
                        ref_result (make_ref (substring (list, m, n - 1)));
                        break;
                    default:
                        error_break1 ("cannot subscript expression: %s",
                                      print_name (list));
		}
	    }
	}
    }
}
